from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import validator
from pydantic.fields import ModelField
from pydantic_forms.core import ReadOnlyField

from gso.products.product_blocks import site as site_pb
from gso.products.product_blocks.site import LatitudeCoordinate, LongitudeCoordinate
from gso.products.product_types.site import Site
from gso.utils.helpers import validate_ipv4_or_ipv6, validate_site_fields_is_unique


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    subscription = Site.from_subscription(subscription_id)

    class ModifySiteForm(FormPage):
        class Config:
            title = "Modify Site"

        site_name: str = ReadOnlyField(subscription.site.site_name)
        site_city: str = subscription.site.site_city
        site_country: str = ReadOnlyField(subscription.site.site_country)
        site_country_code: str = ReadOnlyField(subscription.site.site_country_code)
        site_latitude: LatitudeCoordinate = subscription.site.site_latitude
        site_longitude: LongitudeCoordinate = subscription.site.site_longitude
        site_bgp_community_id: int = subscription.site.site_bgp_community_id
        site_internal_id: int = subscription.site.site_internal_id
        site_tier: site_pb.SiteTier = ReadOnlyField(subscription.site.site_tier)
        site_ts_address: str | None = subscription.site.site_ts_address

        @validator("site_ts_address", allow_reuse=True)
        def validate_ts_address(cls, site_ts_address: str) -> str:
            if site_ts_address and site_ts_address != subscription.site.site_ts_address:
                validate_site_fields_is_unique("site_ts_address", site_ts_address)
                validate_ipv4_or_ipv6(site_ts_address)
            return site_ts_address

        @validator("site_internal_id", "site_bgp_community_id", allow_reuse=True)
        def validate_unique_fields(cls, value: str, field: ModelField) -> str | int:
            if value == getattr(subscription.site, field.name):
                return value
            return validate_site_fields_is_unique(field.name, value)

    user_input = yield ModifySiteForm

    return user_input.dict()


@step("Modify subscription")
def modify_site_subscription(
    subscription: Site,
    site_city: str,
    site_latitude: LatitudeCoordinate,
    site_longitude: LongitudeCoordinate,
    site_bgp_community_id: int,
    site_internal_id: int,
    site_ts_address: str,
) -> State:
    subscription.site.site_city = site_city
    subscription.site.site_latitude = site_latitude
    subscription.site.site_longitude = site_longitude
    subscription.site.site_bgp_community_id = site_bgp_community_id
    subscription.site.site_internal_id = site_internal_id
    subscription.site.site_ts_address = site_ts_address

    subscription.description = f"Site in {site_city}, {subscription.site.site_country}"

    return {"subscription": subscription}


@workflow(
    "Modify Site",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_site() -> StepList:
    return (
        init
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> modify_site_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
